#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Package publish subcommand implementation."""

import copy
import logging
import pathlib
import sys
import typing

import click

from build_harness._utility import (
    CommandArgs,
    command_path,
    report_console_error,
    run_command,
)

from .state import CommandState, ExitState

log = logging.getLogger(__name__)

TWINE_PUBLISH_CMD: CommandArgs = ["twine", "upload", "--non-interactive"]


class PublishingError(Exception):
    """Problem occurred during publishing."""


def _publish_packages(
    venv_path: pathlib.Path,
    is_dryrun: bool,
    password: typing.Optional[str],
    user: typing.Optional[str],
) -> None:
    """
    Publish sdist, wheel packages using twine.

    On dry run no publish occurs, but a summary of packages found that would have
    been published is printed to console.

    Args:
        venv_path: Path to Python virtual environment.
        is_dryrun: Signal if a publish dry run is requested.
        password: Remote repository password.
        user: Remote repository username

    Raises:
        FileNotFoundError: if password file does not exist.
        PublishingError: If package publish exits non-zero.
    """
    this_command = copy.deepcopy(TWINE_PUBLISH_CMD)
    this_command[0] = command_path(venv_path, this_command)
    this_command += ["dist/*.whl", "dist/*.tar.gz"]

    environment_variables = dict()
    if user:
        environment_variables["TWINE_USERNAME"] = user
    if password:
        environment_variables["TWINE_PASSWORD"] = password

    if not is_dryrun:
        result = run_command(
            this_command,
            env=environment_variables,
        )

        if any([x != 0 for x in [result.returncode]]):
            raise PublishingError("twine failed during package publishing.")
    else:
        message = "Dry run, {0}".format(this_command)
        log.warning(message)
        click.echo(message)


@click.command()
@click.pass_context
@click.option(
    "--dryrun",
    default=False,
    help="Run publishing steps without actually publishing artifacts. [default: "
    "do nothing]",
    is_flag=True,
)
@click.option(
    "--password",
    default=None,
    help="PEP-503 server login password",
    type=str,
)
@click.option(
    "--user",
    default=None,
    help="PEP-503 server login user name",
    type=str,
)
def publish(
    ctx: click.Context,
    dryrun: bool,
    password: typing.Optional[str],
    user: typing.Optional[str],
) -> None:
    """Publish project artifacts."""
    try:
        command_state: CommandState = ctx.obj

        _publish_packages(command_state.venv_path, dryrun, password, user)
    except PublishingError as e:
        message = str(e)
        log.exception(message)
        report_console_error(message)
        sys.exit(ExitState.PUBLISHING_FAILED)
    except Exception:
        message = "Unexpected error. Check log for details."
        log.exception(message)
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR)



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

from build_harness.commands._release_id import validate_release_id


def test_public_release():
    expected_release = "3.1.4"
    result = validate_release_id(expected_release)

    assert result.public == expected_release
    assert result.post is None


def test_replace_post():
    expected_release = "3.1.4"
    result = validate_release_id(expected_release)

    assert result.public == expected_release
    assert result.post is None

    v = result.replace(post=5)
    assert v.post == 5


def test_post_release():
    expected_release = "3.1.4.post11"
    result = validate_release_id(expected_release)

    assert result.public == expected_release
    assert result.post == 11


def test_local_dryrun():
    expected_release = "3.1.4.post11+dryrun"
    result = validate_release_id(expected_release)

    assert result.public == "3.1.4.post11"
    assert result.post == 11
    assert result.local == "dryrun"
    assert str(result) == expected_release



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import subprocess

import pytest

from build_harness.commands import main
from tests.ci.support.click_runner import click_runner  # noqa: F401


@pytest.fixture()
def mock_run(mock_sysargv, mocker):
    mock_pytest_result = mocker.create_autospec(subprocess.CompletedProcess)
    mock_pytest_result.returncode = 0
    mock_pytest_result.stdout = ""
    this_run = mocker.patch(
        "build_harness.commands.publishing.run_command",
        return_value=mock_pytest_result,
    )

    return this_run


class TestPublish:
    def test_default(self, click_runner, mocker, mock_run):
        result = click_runner.invoke(
            main,
            [
                "publish",
                "--user",
                "username",
                "--password",
                "password",
            ],
        )

        assert result.exit_code == 0
        mock_run.assert_has_calls(
            [
                mocker.call(
                    [
                        "/some/path/twine",
                        "upload",
                        "--non-interactive",
                        "dist/*.whl",
                        "dist/*.tar.gz",
                    ],
                    env={
                        "TWINE_PASSWORD": "password",
                        "TWINE_USERNAME": "username",
                    },
                ),
            ]
        )



#
#  Copyright (c) 2021 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Release flow command entry point."""

import logging
import os
import pathlib
import shutil
import sys
import typing

import click

from build_harness._utility import report_console_error
from build_harness._version import acquire_version
from build_harness.tools.git import (
    DEFAULT_DEFAULT_BRANCH_NAME,
    GitRepoError,
    get_tag_data,
)

from ._declarations import DEFAULT_PROJECT_PATH
from ._release_id import InvalidReleaseId, validate_release_id
from .state import CommandState, ExitState

log = logging.getLogger(__name__)


class GitNotFoundError(Exception):
    """Git executable not found error."""


def _validate_git(optional_git: typing.Optional[str]) -> None:
    """
    Validate that git is installed or the user has provided a valid git path.

    Args:
        optional_git: Path to git executable.

    Raises:
        GitNotFoundError: If git executable not found.
    """
    system_git = shutil.which("git")
    if optional_git:
        system_git = optional_git

    if not system_git:
        raise GitNotFoundError("Git not available on users PATH")
    if not os.path.exists(system_git):
        raise GitNotFoundError("User specified git invalid, {0}".format(system_git))


@click.command()
@click.version_option(version=acquire_version())
@click.option(
    "--default-branch",
    default=DEFAULT_DEFAULT_BRANCH_NAME,
    help="Git default/main branch name.",
    show_default=True,
    type=str,
)
@click.option(
    "--git",
    default=None,
    help="Git executable. [default: users PATH]",
    type=str,
)
@click.option(
    "--project",
    default=DEFAULT_PROJECT_PATH,
    help="Git project directory.",
    show_default=True,
    type=str,
)
def release_flow_main(
    default_branch: str, git: typing.Optional[str], project: str
) -> None:
    """
    Construct a release id from git tag and commit history.

    If HEAD commit is tagged then use the tag as the release id. Otherwise try to
    count the number of commits since the most recent tag, with a PEP-440 constructed
    release id of the form ``<tag>.post<N>``. Note that this implies that the tag
    must be PEP-440 compliant.

    If there are no tags in commit history then a default tag of ``0.0.0`` is
    inferred and the number of commits counted since the root (first) commit in
    commit history.

    The count method is slightly different when there are no tags in commit history
    so it is recommended to tag the root commit with ``0.0.0`` so that the count
    method is always the same.
    """
    try:
        state = CommandState(
            project_path=pathlib.Path(project),
            venv_path=pathlib.Path(sys.argv[0]).parent.absolute(),
        )

        _validate_git(git)

        tag_data = get_tag_data(state.project_path, default_branch)
        tag_version = validate_release_id(tag_data.tag)
        if tag_data.offset:
            updated_version = tag_version.replace(post=int(tag_data.offset))
            release_id = str(updated_version)
        else:
            release_id = str(tag_version)

        click.echo(release_id, err=False, nl=False)
    except InvalidReleaseId as e:
        report_console_error(str(e))
        sys.exit(ExitState.BAD_VERSION)
    except GitRepoError as e:
        report_console_error(str(e))
        sys.exit(ExitState.BAD_REPO)
    except GitNotFoundError as e:
        report_console_error(str(e))
        click.echo("Git must be installed and configured to use this utility.")
        click.echo("In Linux, something like this is necessary:")
        click.echo("    sudo apt install -y git || sudo yum install git")
        click.echo('    git config user.name "Your Name"')
        click.echo('    git config user.email "you@example.com"')
        sys.exit(ExitState.BAD_GIT_EXE)
    except Exception:
        message = "Unexpected error. Check log for details."
        log.exception(message)
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Formatting subcommand implementation."""

import copy
import logging
import pathlib
import sys

import click

from build_harness._utility import (
    CommandArgs,
    command_path,
    report_console_error,
    run_command,
)

from .state import CommandState, ExitState

BLACK_FORMAT_CMD: CommandArgs = ["black", "."]
ISORT_FORMAT_CMD: CommandArgs = ["isort", "--profile", "black", "."]

BLACK_CHECK_CMD: CommandArgs = ["black", "--check", "."]
ISORT_CHECK_CMD: CommandArgs = ["isort", "--profile", "black", "--diff", "."]

log = logging.getLogger(__name__)


class FormattingError(Exception):
    """Problem occurred during code formatting."""


class IsortCheckError(Exception):
    """The isort formatting check failed."""


class BlackCheckError(Exception):
    """The black formatting check failed."""


def _apply_formatting(venv_path: pathlib.Path) -> None:
    """
    Apply formatting to source code.

    Args:
        venv_path: Path to Python virtual environment.

    Raises:
        FormattingError: If either isort or black exit non-zero.
    """
    isort_cmd = copy.deepcopy(ISORT_FORMAT_CMD)
    isort_cmd[0] = command_path(venv_path, isort_cmd)
    isort_result = run_command(isort_cmd)

    black_cmd = copy.deepcopy(BLACK_FORMAT_CMD)
    black_cmd[0] = command_path(venv_path, black_cmd)
    black_result = run_command(black_cmd)

    if any([x != 0 for x in [isort_result.returncode, black_result.returncode]]):
        raise FormattingError("isort and/or black failed during formatting.")


def _check_formatting(venv_path: pathlib.Path) -> None:
    """
    Check if any formatting changes are needed on the code base.

    Args:
        venv_path: Path to Python virtual environment.

    Raises:
        BlackCheckError: If black formatting check fails.
        IsortCheckError: If isort formatting check fails.
    """
    isort_cmd = copy.deepcopy(ISORT_CHECK_CMD)
    isort_cmd[0] = command_path(venv_path, isort_cmd)
    isort_result = run_command(isort_cmd)
    if isort_result.returncode != 0:
        raise IsortCheckError("isort check failed")

    black_cmd = copy.deepcopy(BLACK_CHECK_CMD)
    black_cmd[0] = command_path(venv_path, black_cmd)
    black_result = run_command(black_cmd)
    if black_result.returncode != 0:
        raise BlackCheckError("black check failed")


@click.command()
@click.pass_context
@click.option(
    "--check",
    help="Don't apply formatting, just check if formatting is needed.",
    is_flag=True,
)
def formatting(ctx: click.Context, check: bool) -> None:
    """
    Run suite of Python code formatting.

    The suite currently comprises "isort and "black". isort is run before black to
    resolve cases where isort and black compete over formatting changes.
    """
    try:
        command_state: CommandState = ctx.obj
        if not check:
            _apply_formatting(command_state.venv_path)
        else:
            _check_formatting(command_state.venv_path)
    except FormattingError as e:
        report_console_error(str(e))
        sys.exit(ExitState.FORMATTING_FAILED)
    except BlackCheckError as e:
        report_console_error(str(e))
        sys.exit(ExitState.BLACK_CHECK_FAILED)
    except IsortCheckError as e:
        report_console_error(str(e))
        sys.exit(ExitState.ISORT_CHECK_FAILED)
    except Exception:
        message = "Unexpected error. Check log for details."
        log.exception(message)
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

"""Static analysis subcommand implementation."""

import logging
import pathlib
import sys

import click

from build_harness._project import acquire_source_dir
from build_harness._utility import (
    CommandArgs,
    command_path,
    report_console_error,
    run_command,
)

from .code_style import FormattingError, _apply_formatting
from .state import CommandState, ExitState

log = logging.getLogger(__name__)


FLAKE8_FORMAT_CMD: CommandArgs = ["flake8", "."]
MYPY_FORMAT_CMD: CommandArgs = ["mypy", "--show-error-codes", "."]
PYDOCSTYLE_FORMAT_CMD: CommandArgs = ["pydocstyle", "."]

VALID_ANALYSIS_OPTIONS = [
    "all",
    "flake8",
    "mypy",
    "pydocstyle",
]


class StaticAnalysisError(Exception):
    """Problem occurred during static analysis."""


class Flake8Error(StaticAnalysisError):
    """Problem occurred with flake8 analysis."""


class MypyError(StaticAnalysisError):
    """Problem occurred with mypy analysis."""


class PydocstyleError(StaticAnalysisError):
    """Problem occurred with pydocstyle anslysis."""


def _run_flake8(source_path: pathlib.Path, venv_path: pathlib.Path) -> None:
    """
    Run flake8 analysis on the specified source directory.

    Args:
        source_path: Path to project source directory.
        venv_path: Path to Python virtual environment.

    Raises:
        StaticAnalysisError: If flake8 analysis fails.
    """
    FLAKE8_FORMAT_CMD[0] = command_path(venv_path, FLAKE8_FORMAT_CMD)
    FLAKE8_FORMAT_CMD[-1] = str(source_path)
    result = run_command(FLAKE8_FORMAT_CMD)

    if result.returncode != 0:
        raise Flake8Error("flake8 analysis failed")


def _run_mypy(source_path: pathlib.Path, venv_path: pathlib.Path) -> None:
    """
    Run mypy analysis on the specified source directory.

    Args:
        source_path: Path to project source directory.
        venv_path: Path to Python virtual environment.

    Raises:
        StaticAnalysisError: If mypy analysis fails.
    """
    MYPY_FORMAT_CMD[0] = command_path(venv_path, MYPY_FORMAT_CMD)
    MYPY_FORMAT_CMD[-1] = str(source_path)
    result = run_command(MYPY_FORMAT_CMD)

    if result.returncode != 0:
        raise MypyError("mypy analysis failed")


def _run_pydocstyle(source_path: pathlib.Path, venv_path: pathlib.Path) -> None:
    """
    Run pydocstyle analysis on the specified source directory.

    Args:
        source_path: Path to project source directory.
        venv_path: Path to Python virtual environment.

    Raises:
        StaticAnalysisError: If pydocstyle analysis fails.
    """
    PYDOCSTYLE_FORMAT_CMD[0] = command_path(venv_path, PYDOCSTYLE_FORMAT_CMD)
    PYDOCSTYLE_FORMAT_CMD[-1] = str(source_path)
    result = run_command(PYDOCSTYLE_FORMAT_CMD)

    if result.returncode != 0:
        raise PydocstyleError("pydocstyle analysis failed")


def _apply_analysis(
    analysis_type: str, project_path: pathlib.Path, venv_path: pathlib.Path
) -> None:
    """
    Apply the specified analysis.

    Args:
        project_path: Path to project root directory.
        venv_path: Path to Python virtual environment.
    """
    source_dir = acquire_source_dir(project_path)

    analysis_methods = {
        "flake8": _run_flake8,
        "mypy": _run_mypy,
        "pydocstyle": _run_pydocstyle,
    }
    if analysis_type != "all":
        analysis_methods[analysis_type](project_path / source_dir, venv_path)
    else:
        for this_method in analysis_methods.values():
            this_method(project_path / source_dir, venv_path)


@click.command()
@click.option(
    "--analysis",
    default="all",
    help="Analysis type to apply (default all)",
    type=click.Choice(VALID_ANALYSIS_OPTIONS, case_sensitive=True),
)
@click.pass_context
def static_analysis(ctx: click.Context, analysis: str) -> None:
    """
    Run suite of Python static analysis.

    The suite currently comprises "flake8", "pydocstyle" and "mypy". Formatting is
    run before the static analysis to eliminate formatting complaints from static
    analysis tools.
    """
    try:
        command_state: CommandState = ctx.obj

        _apply_formatting(command_state.venv_path)
        _apply_analysis(analysis, command_state.project_path, command_state.venv_path)
    except FormattingError as e:
        report_console_error(str(e))
        sys.exit(ExitState.FORMATTING_FAILED)
    except Flake8Error as e:
        report_console_error(str(e))
        sys.exit(ExitState.FLAKE8_FAILED)
    except MypyError as e:
        report_console_error(str(e))
        sys.exit(ExitState.MYPY_FAILED)
    except PydocstyleError as e:
        report_console_error(str(e))
        sys.exit(ExitState.PYDOCSTYLE_FAILED)
    except Exception:
        message = "Unexpected error. Check log for details."
        log.exception(message)
        report_console_error(message)
        sys.exit(ExitState.UNKNOWN_ERROR)



#
#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of build_harness.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import features.steps.structure.bootstrap_steps
import features.steps.structure.common_steps
import features.steps.structure.package_steps
import features.steps.structure.release_flow_steps



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['build_harness',
 'build_harness.commands',
 'build_harness.commands.bdd',
 'build_harness.tools']

package_data = \
{'': ['*']}

install_requires = \
['behave == 1.2.6',
 'black == 20.8b1',
 'click == 7.1.2',
 'flake8 == 3.8.4',
 'flake8-annotations == 2.4.1',
 'flit == 3.0.0',
 'gitpython == 3.1.14',
 'isort == 5.5.4',
 'mypy == 0.782',
 'packaging == 20.4',
 'parver == 0.3.1',
 'pip == 20.2.4',
 'pydocstyle == 5.1.1',
 'pygments == 2.7.1',
 'toml == 0.10.1',
 'twine == 3.3.0']

extras_require = \
{'dev': ['mutmut == 2.1.0', 'pre_commit == 2.7.1'],
 'doc': ['sphinx == 3.2.1', 'sphinx_rtd_theme == 0.5.0'],
 'test': ['pytest == 6.1.1', 'pytest-cov == 2.10.1', 'pytest-mock == 2.0.0']}

entry_points = \
{'console_scripts': ['build-harness = '
                     'build_harness.entrypoint:build_harness_entry',
                     'release-flow = '
                     'build_harness.entrypoint:release_flow_entry']}

setup(name='build_harness',
      version='0.1.1',
      description='CI build harness embodying best practices for Python projects.',
      author='Russell Smiley',
      author_email='im.russell.smiley@gmail.com',
      url='https://gitlab.com/ci-cd-devops/build_harness',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.7',
     )
